{ ****************************************************************************
  `X' - eXtract byte sequences (main executable)

  Written by Robert Riebisch <rr@bttr-software.de>
  for BTTR Software

  Copyright (c) 1997-2003 BTTR Software
  All rights reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the `MODIFIED' BSD LICENSE.

  See `legal.txt' for details.
  **************************************************************************** }

{.$DEFINE DEBUG}

{$IFDEF DEBUG}
  {$A+,B-,D+,E-,F-,I-,L+,N-,O-,R+,S+,V+}
  {$M 2048,65535,65535}
{$ELSE}
  {$A+,B-,D-,E-,F-,I-,L-,N-,O-,R-,S-,V+}
  {$M 2048,65535,65535}
{$ENDIF}

program x;

uses
  strconv;

const
  BUFSIZE = 65535;  { size of read/write buffer }

var
  parcount: Word;  { ParamCount holder }

  inf,         { input file variable }
  outf: file;  { output file variable }

  infname,           { input file name }
  outfname: string;  { output file name }

  infsize: Longint;  { input file size }

  valc: Integer;  { Val() return code }

  xstart,         { starting point }
  xlen,           { no. of bytes }
  maxlen,         { maximum no. of bytes }
  pc,             { percentage (status of dissection) }
  thresh,         { threshold (tens mark) }
  i, j: Longint;  { loop counters }

  buf: Pointer;  { pointer to memory buffer on heap }

{ get DOS version }
function osmajor: Byte;
inline(
  $b4/$30/  { mov ah, $30  ; function `GET DOS VERSION' }
  $cd/$21   { int $21      ; call DOS }
);

{ canonicalize filename or path }
procedure truename(var fname);
begin
  { convert fname to ASCIIZ string }
  strpastoc(fname);

  { call DOS `TRUENAME' function }
  inline(
    $1e/             { push ds        ; save DS }
    $c5/$76/<fname/  { lds si, fname  ; load address of ASCIIZ filename or path }
    $c4/$7e/<fname/  { les di, fname  ; load address of 128-byte buffer for canonicalized name }
    $b4/$60/         { mov ah, $60    ; function `TRUENAME' }
    $cd/$21/         { int $21        ; call DOS }
    $1f              { pop ds         ; restore DS }
  );

  { convert fname back to Pascal string }
  strctopas(fname);
end;

{ add thousands separators to a Longint value, e.g., modify 1000 to 1,000 }
function addcommas(l: Longint): string;
var
  i: Integer;  { loop counter }

  s: string[20];  { string-converted value }
begin
  Str(l, s);
  i := Length(s) - 2;
  while i > 1 do
  begin
    Insert(',', s, i);
    Dec(i, 3);
  end;
  addcommas := s;
end;

{ returns the high-order Integer of a Longint }
function longhi(l: Longint): Integer;
inline(
  $58/  { pop ax  ; throw away low word }
  $58   { pop ax  ; get high word }
);

{ compute Trunc(100 * x / y) w/o overflow or using `/' }
function truncpc(x, y: Longint): Integer;
begin
  if longhi(y) < $100 then
    truncpc := (100 * x) div y
  else
    truncpc := x div (y div 100);
end;

begin
  { name and license of program }
  Write('X Version 28-APR-2003'^M^J +
        'Copyright (c) 1997-2003 BTTR Software'^M^J +
        '[Under `MODIFIED'' BSD LICENSE]'^M^J +
        ^M^J);

  { check DOS version needed for truename() }
  if osmajor < 3 then
  begin
    Write('Error: DOS 3.0 or later required.'^M^J);
    Halt(1);
  end;

  { output usage if no parameters or command line contains `/?' }
  parcount := ParamCount;
  if (parcount = 0) or (Pos('/?', string(Ptr(PrefixSeg, $80)^)) > 0) then
  begin
    Write('Usage: x <infile> <outfile> <offset> [<length>]'^M^J +
          ^M^J +
          '  infile  = input file name (path optional)'^M^J +
          '  outfile = output file name (path optional)'^M^J +
          '  offset  = starting point (in infile) for dissection'^M^J +
          '  length  = # of bytes to extract to outfile (optional)'^M^J);
    Halt;
  end;

  { check number of parameters }
  if (parcount < 3) or (parcount > 4) then
  begin
    Write('Error: Invalid number of parameters.'^M^J);
    Halt(1);
  end;

  { expand filenames with drive letter and absolute path }
  infname := ParamStr(1);
  truename(infname);

  outfname := ParamStr(2);
  truename(outfname);

  { check filenames }
  if outfname = infname then
  begin
    Write('Error: Output file must not match input file.'^M^J);
    Halt(1);
  end;

  { open input file }
  FileMode := 0;
  Assign(inf, infname);
  Reset(inf, 1);
  if IOResult <> 0 then
  begin
    Write('Error: Unable to open input file.'^M^J);
    Halt(1);
  end;

  { get size of input file for later calculations }
  infsize := FileSize(inf);

  { convert and check offset parameter }
  Val(ParamStr(3), xstart, valc);
  if (valc <> 0) or (xstart < 0) or (xstart >= infsize) then
  begin
    Write('Error: Invalid offset specified.'^M^J);
    Halt(1);
  end;

  { calculate maximum length to avoid seeks beyond end of input file }
  maxlen := infsize - xstart;

  { extract to end of input file? }
  if parcount = 3 then
    xlen := maxlen
  else
  begin
    { convert and check length parameter }
    Val(ParamStr(4), xlen, valc);
    if (valc <> 0) or (xlen < 1) then
    begin
      Write('Error: Invalid length specified.'^M^J);
      Halt(1);
    end;

    { cut length down }
    if xlen > maxlen then
      xlen := maxlen;
  end;

  { open output file }
  FileMode := 1;
  Assign(outf, outfname);
  Rewrite(outf, 1);
  if IOResult <> 0 then
  begin
    Write('Error: Unable to open output file.'^M^J);
    Halt(1);
  end;

  { seek to offset in input file }
  Seek(inf, xstart);
  if IOResult <> 0 then
  begin
    Write('Error: Unable to reach offset.'^M^J);
    Halt(1);
  end;

  { check available heap memory }
  if MaxAvail < BUFSIZE then
  begin
    Write('Error: Not enough memory to allocate buffer.'^M^J);
    Halt(1);
  end;

  { allocate read/write buffer }
  GetMem(buf, BUFSIZE);

  { display # of bytes to be extracted }
  Write('Extracting ', addcommas(xlen), ' byte(s)...'^M^J +
        ^M^J);

  { main `copy' routine }
  thresh := 0;
  i := 0;
  while i < xlen do
  begin
    { trim bytes to be read to trailing bytes }
    j := xlen - i;
    if j > BUFSIZE then
      j := BUFSIZE;

    { read input file into buffer }
    BlockRead(inf, buf^, j);
    if IOResult <> 0 then
    begin
      Write('Error: Unable to read from input file.'^M^J);
      Halt(1);
    end;

    { write output file from buffer }
    BlockWrite(outf, buf^, j);
    if IOResult <> 0 then
    begin
      Write('Error: Unable to write to output file.'^M^J);
      Halt(1);
    end;

    { increase total bytes written }
    Inc(i, j);

    { calculate percentage and display tens mark if threshold reached }
    pc := truncpc(i, xlen);
    while pc >= thresh do
    begin
      Write(thresh);
      if thresh < 100 then
        Write('%..');
      Inc(thresh, 10);
    end;
  end;

  { final percent sign and CR/LF }
  Write('%'^M^J);

  { free buffer on heap }
  FreeMem(buf, BUFSIZE);

  { close files }
  Close(inf);
  Close(outf);
end.